/***
*		MT_util.c -- utility-type functions to link with MakeTuning
*		 1994 Another Option   All rights reserved.
***/

/*
============================================================================
|    DATE   | WHAT
============================================================================
  31 Jan 94 | extracted from maketuning.c because it took too bleeding long
            | to compile otherwise
  19 Feb 94 | modified cents_to_pitch: out of range notes now marked unused
            | instead of folding into midi range.
*/

#include <exec/exec.h>
#include <proto/exec.h>
#include <intuition/intuition.h>
#include <proto/intuition.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "bars.h"
#include "barsMacros.h"
#include "bplib.h"
#include "tuning.h"
#include "proto/mt_util_protos.h"

/*** defined in maketuning.c ***/
extern struct MakeTuneAccessory MakeTune;
extern struct Tuning pitchtable;
extern char displaystart;
extern char edited;
extern char used;
extern UBYTE root_edited;

extern UBYTE *WTitle;

extern void __regargs _CXBRK(void);
extern void _XCEXIT(long );

extern struct Functions *functions;


/*==========================================================================
* stuffoctavetext
*
*  INPUT: pointer to gadget, octave value
* RETURN: none
*
* EFFECT: sprintf()s octave value into gadget's text, and adjusts text's
*         LeftEdge accordingly for 1 or 2 characters of text
*=========================================================================*/

void stuffoctavetext(g,oct)

struct Gadget *g;
char oct;

{
	sprintf((char *)(g->GadgetText->IText),"%*d",(oct > 9) ? 2 : 1, oct);
	g->GadgetText->LeftEdge = (oct > 9) ? 3 : 7;
}


/*** a length for window title/filename display ***/
#define WTITLE_LEN 25

/*============================================================================
* changename
*
*  INPUT: pointer to name
* RETURN: none
*
* EFFECT: sets the pitchtable name. if a name is passed it is used.
*         if null is passed, the current pitchtable.name is redisplayed.
*===========================================================================*/

void changename(const char * name)

{
	static char WTitle[WTITLE_LEN+1] = { 0 };
//		"Tonleitermeister\0" };

	if (name) {
		if (pitchtable.name) functions->freestring(pitchtable.name);
#pragma msg 104 ignore push
		pitchtable.name = functions->allocstring(name);
#pragma msg 104 ignore pop
	}
	if (pitchtable.name) {
		sprintf(WTitle,"%-20s",functions->stringtext(pitchtable.name));
		functions->EmbossedWindowTitle(MakeTune.acc.window,WTitle);
	}
}

/*==========================================================================
* round
*
*  INPUT: double
* RETURN: long
*
* EFFECT: returns the least erroneous long integer
*=========================================================================*/

long round(double n)

{
	double nc, nf;

	nc = ceil(n);
	nf = floor(n);
	return( (nc-n) <= (n-nf) ? (long)nc : (long)nf );
}


/**************************** some conversions **************************/

/* gets interval in cents of p from root */
#define pitchtocents(p) ( ( (p) - ptroot ) * 100 )


/*?? why aren't these conversions macros? ??*/
/*==========================================================================
* bend_to_cents
*
*  INPUT: bytes of a MIDI bend word
* RETURN: cents
*
* EFFECT: converts the bend to equivalent in cents
*=========================================================================*/

long bend_to_cents(UBYTE lsb, UBYTE msb)

{
//	return(round((MIDIWord(msb,lsb)-MIDI_BendCntr)*100.0/MIDI_BendCntr));
	return((long)((MIDIWord(msb,lsb)-MIDI_BendCntr)*100.0/MIDI_BendCntr));
}


/*==========================================================================
* cents_to_bend
*
*  INPUT: cents
* RETURN: bend
*
* EFFECT: converts cents to equivalent 14-bit bend
*=========================================================================*/

UWORD cents_to_bend(long cents)

{
	return( (UWORD)round(MIDI_BendCntr * cents / 100.0 + MIDI_BendCntr) );
}

/* log(2) */
#define LOG2 0.69314718

/*============================================================================
* cents_to_ratio
*
*  INPUT: cents
* RETURN: ratio
*
* EFFECT: converts a span of cents from a root to a ratio to the root
*===========================================================================*/

double cents_to_ratio(double cents)

{
	return( exp((cents * LOG2)/1200.0) );
}
	
/*============================================================================
* ratio_to_cents
*
*  INPUT: ratio in decimal form
* RETURN: cents
*
* EFFECT: converts a ratio to the equivalent span in cents
*===========================================================================*/

double ratio_to_cents(double ratio)

{
	return( (1200.0 * log(ratio)) / LOG2 );
}


/* get i-th pitchtable midinote */
#define ptnote(i) (pitchtable.notes[(i)].midinote & 0x7F)

/* calculate the root's note number */
#define ptroot ( pitchtable.rootpitch + pitchtable.rootoctave * 12 )

/*============================================================================
* cents_to_pitch
*
*  INPUT: cents, index into pitchtable
* RETURN: none
*
* EFFECT: converts.  if the resulting pitch is greater than 127 or less than
*         0, /*it is transposed NOT*/ marked unused by setting the high bit in
*         bendlsb.  if cents is negative, it defines an interval below the 
*         rootpitch/rootoctave.
*===========================================================================*/

void cents_to_pitch(double cents, long index)

{
	long pitch = round(cents/100.0);
	long bend = cents_to_bend( round(cents-(pitch * 100)) );

	pitchtable.notes[index].bendlsb = MIDILoByte(bend);
	pitchtable.notes[index].bendmsb = MIDIHiByte(bend);
	pitch += ptroot;
/*
	while (pitch > 127) pitch -= 12;
	while (pitch < 0) pitch += 12;
*/
	/* mark out of range midinotes as unused, but with tolerable note values */
	if (pitch > 127) {
		pitch=0xFF;    /* not used 127 */
		pitchtable.notes[index].bendlsb |= 0x80;
	}
	else if (pitch < 0) {
		pitch=0x80; /* not use 0 */
		pitchtable.notes[index].bendlsb |= 0x80;
	}
	pitchtable.notes[index].midinote = (char)pitch;
}

/*==========================================================================
* RethinkPitchTable
*
*  INPUT: none
* RETURN: none
*
* EFFECT: recalculates the entire pitch table based on the current ratios
*=========================================================================*/

void RethinkPitchTable()

{
	register ULONG i;

	for (i=0 ; i<128 ; i++) {
		cents_to_pitch(ratio_to_cents(atof(&pitchtable.ratios[i][0])),i);
		if (i != ptroot)
			pitchtable.notes[i].midinote |= 0x80; /* not edited */
	}
}

/*==========================================================================
* UpdateRatio
*
*  INPUT: index, ratio
* RETURN: none
*
* EFFECT: updates pitchtable.ratios[index] string.  if ratio>0.0, just do a
*         sprintf(). if ratio<=0.0, calculate a new ratio based on the 
*         midinote and bend entries in the pitchtable.
*=========================================================================*/

void UpdateRatio(ULONG index, double ratio)

{
	if (ratio > 0.0) {
		sprintf(&pitchtable.ratios[index][0],RATFORM,ratio);
	}
	else {
		ratio = (double)pitchtocents(ptnote(index));
		ratio += (double)bend_to_cents(pitchtable.notes[index].bendlsb & 0x7F,
		                             pitchtable.notes[index].bendmsb);
		sprintf(&pitchtable.ratios[index][0],RATFORM,cents_to_ratio(ratio));
	}
}

/*==========================================================================
* RethinkRatios
*
*  INPUT: none
* RETURN: none
*
* EFFECT: calls UpdateRatio() several times
*=========================================================================*/

void RethinkRatios()

{
	UBYTE i;

	for ( i=0 ; i<128 ; i++ ) UpdateRatio(i,-1.0);
}

/*==========================================================================
* RecalcPitchTable
*
*  INPUT: divisisons per octave
* RETURN: none
*
* EFFECT: makes a new equally-tempered tuning
*=========================================================================*/

void RecalcPitchTable(ULONG divisions, long root)

{
	long i;
	double step = 1200.0/divisions; /* cents between scale degrees */

 	for ( i=0 ; i<128 ; i++ ) {
		cents_to_pitch(step*(i-root),i);
		pitchtable.notes[i].midinote |= 0x80;
		UpdateRatio(i,-1.0);
	}
}

/*==========================================================================
* initpitchtable
*
*  INPUT: none
* RETURN: none
*
* EFFECT: set pitchtable[] to defaults, changename() to default NONAME.
*
*   NOTE: The high bit of each note number is set as a flag to indicate that
*         a note number has not been edited.
*=========================================================================*/

/* tuning with no name */
#define NONAME "Unnamed Tuning"


void initpitchtable(void)

{
	UBYTE i;
	
	if (used && pitchtable.name) functions->freestring(pitchtable.name);
	memset((char *)&pitchtable,0,sizeof(struct Tuning));
	pitchtable.divisions = 12;
	pitchtable.rootoctave = 5;
	for (i=0 ; i<128 ; i++) {
		pitchtable.notes[i].midinote = i | 0x80; /* flag unset pitch */
//		pitchtable.notes[i].bendlsb = 0;		/* reset to MIDI_BendCntr whilst */
		pitchtable.notes[i].bendmsb = 0x40;	/* dispensing with bit twiddling */
	}
	RethinkRatios();
	pitchtable.notes[60].midinote &= 0x7F; /* root always edited */
	if (MakeTune.acc.window) changename(NONAME);
}
