/***
* 		tuning.h -- definitions for alternate tunings for b&p
*		 1991 - 1994 Another Option
***/

/*
============================================================================
|    DATE    | WHAT
============================================================================
	 9 Sep 91 | created
	 2 Oct 91 | added ratios to struct Tuning to avoid confusion when
	            toggling between bends and ratios in maketuning.c
	14 Oct 91 | updated file format documentation
	28 Jan 94 | added port name and command defines; expanded accessory struct
	19 Feb 94 | atun tool now loads dopt node of a file for export to accessory
*/

#ifndef TUNING_H
#define TUNING_H

#include "bars.h"

/*
#ifndef BARS_MACROS_H
 #include "barsmacros.h"
#endif
*/

/*** Alternate Tuning tool stuff ***/
#define ATUN_TOOLID 0x4154554E

struct TuneTool {
	struct Tool tool;
	struct Tuning *tuning; /* the tuning in use by this tool */
	char bendlsb;          /* last bend value sent */
	char bendmsb;
};

struct TuningToolMaster {
	struct ToolMaster master;
	struct MsgPort *mp;  /* all instances of tool share this port */
	struct Tuning *list; /* list of loaded Tunings */
	ULONG mplock;        /* a tool instance has control of the msgport */
};


/*** this expanded accessory structure is so the tuning tool can check to see
 *   if Tonleitermeister is already opened and running or not by checking the
 *   flag.
 ***/
struct MakeTuneAccessory {
	struct Accessory acc;
	struct MsgPort *mp;  /* public port for talking to AltTuning tool */
	struct Task *task;   /* task created in open() to monitor the msg port */
	long running;        /* flag for whether or not the accessory is running */
};

#define EVENT_TUNING 66 /* event type -- this'll do for the time being */

/* an event in case this is ever integrated with the rest of b&p (hint ;) */
/* to support dynamic tunings */
struct TuningEvent {
	struct Event *next;
	long time;
	char type;
	char pad[3];
	unsigned long tuningid;
	struct Tuning *tuning;
};

/* each instance of a Tuning contains an array of midi note numbers and 
   associated pitchbend values.  The indices of this array map to received
	midi note numbers.  For each note number received, a corresponding note
	number plus pitch bend is output to yield the defined tuning system.
	if the high bit of a midinote is set, it indicates the note is not
	defined, and should be ignored.  This is to support tuning systems that
	have fewer than 128 pitches.  In maketuning.c, the Use Edited gadget
	determines whether or not this feature is used.
*/

/* number of characters for ascii ratios - including NULL */
#define RATCHARS 12
/* format for ratios in file and in display */
#define RATFORM "%11.6f"

struct Tuning {
	struct Tuning *next;
	struct String *name; /*!! this should have been a char[] */
	char rootpitch;
	char rootoctave;
	char divisions;      /* notes per octave */
	char dopt_flags;     /* display option flags */
	struct PitchDef {
		char midinote;
		char bendlsb;     /* these are stored in MIDI format (14-bit math) */
		char bendmsb;     /* so they are ready to transmit */
	} notes[128];
	char ratios[128][RATCHARS]; /* ascii ratio strings */
};

/* dopt_flags bits */
#define DOPT_NUMS 1
#define DOPT_RATIOS 2
#define DOPT_MASK 4

/* message port names */
#define TOOL_PORT_NAME "BPP_ATun"
#define ACCESS_PORT_NAME "BPP_Toon"
#define ACCESS_TASK_NAME "BPP_ToonEar"

/* message structure for communication between AltTuning & Tonleitermeister */
struct TuneMessage {
	struct Message message;
	struct Tuning *tuning;
	UBYTE command;
	UBYTE pad[3];
};

/*** TuneMessage commands ***/
/* tool wants a pointer to a tuning */
#define CMD_GIMMEE 1
/* accessory provides a pointer to a tuning */
#define CMD_TAKEIT 2
/* tool says edit this tuning, you accessory you */
#define CMD_EDIT   3
/* got what you passed me; carry on */
#define CMD_ACK    4
/* it was not possibly to comply with the previous request, so forget it */
#define CMD_BITEME 5
/* time for Tonleitermeister's message port task to remove itself */
#define CMD_DIE 6

/*** a Tuning without ratios -- used in tuning.c ***/
#define TUNING_NORAT_SIZE ( sizeof(struct Tuning)-(128*RATCHARS) )

/*** file chunk ids ***/
#define ID_TUNE 0x544F4F4E  //"TOON"
#define ID_DOPT 0x444F5054  //"DOPT"
#define ID_NAME 0x4E414D45  //"NAME"
#define ID_PTAB 0x50544142  //"PTAB"
#define ID_RATS 0x52415453  //"RATS"

/********************************************************************/
/************************* tuning file format ***********************

	long 'TOON'                       4
	long fileSize                     4
	long 'DOPT'                       4
	long doptSize                     4
	long dopts                        4
	long 'NAME'                       4
	long nameSize                     4
	char name[]                       nameSize
	long 'PTAB'                       4
	long ptabSize                     4 (128*sizeof(struct PitchDef)+2)
	char rootpitch                    1
	char rootoctave                   1
	struct PitchDef pitchtable[128]   ptabSize-2 (384)
  ---- optional ----
	long 'RATS'                       4
	long ratSize                      4
	char ratios[ptabSize][RATCHARS]   ratSize (1536)
  ------------------
*********************************************************************/

/* size of full-blown file (less nameSize, which is variable) */
#define FILESIZE_FIXED 1966
/* size of minimal file (less name, which is variable) */
#define FILESIZE_NORAT 422

/********************************************************************/
/************************** b&p save format *************************

	long 'TOON'                       4
	long fileSize                     4
	long 'DOPT'                       4
	long doptSize                     4
	long dopts                        4
	long 'NAME'                       4
	long nameSize                     4
	char name[]                       nameSize
	long 'PTAB'                       4
	long ptabSize                     4 (128*sizeof(struct PitchDef)+2)
	char rootpitch                    1 ! not used, but makes reading and
	char rootoctave                   1 ! writing simpler
	struct PitchDef pitchtable[128]   ptabSize (384)

*********************************************************************/

/* bytes in a save to a song file, less the length of name[] */
#define TOONDATASIZE FILESIZE_NORAT

#endif
