/***
*		tuning.c -- alternate tuning tool for b&p
*		 1991 - 1999  Another Option
***/

/*
============================================================================
|    DATE   | WHAT
============================================================================
	 9 Sep 91 | created
	22 Sep 91 | added load,select,remove tuning files, and sundry
	            supporting functions
	28 Oct 91 | added new load and tuning file format
	~9 Nov 91 | tweeked ui; added loadtool(), savetool(), added bend
	          | remembery in proccessevent()
	22 Apr 92 | tweeked
   26 Jan 93 | massaged for sas/c v6.0
   11 Feb 93 | massaged for new bplib.lib
	11 Nov 93 | massaged for new window title form
	-----------------------------------------------------------------
	26 Jan 94 | new ui with menus and call to tonleitermeister on edit (i hope)
	 1 Feb 94 | finally, can import a tuning from tonleitermeister.
	17 Feb 94 | now, how about exporting...
	18 Feb 94 | ok, you got it.
	24 Feb 94 | added support for skipping invalid notes on bendlsb & 0x80 in
	          | addition to nonedited notes.
	31 Aug 95 | fixed the bullshit screw ups from last time.
	 2 Sep 95 | maybe for real this time.
	 1 Mar 99 | you left out some __stdargs; replaced BozoRequest() with msgbox()
*/

#define DBUG 0

#define THIS_TOOL_NAME "ATUN"

#include <exec/types.h>
#include <exec/exec.h>
#include <exec/memory.h>
#include <exec/ports.h>
#include <libraries/dos.h>
#include <string.h>
#include <intuition/intuition.h>
#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/graphics.h>
#include <proto/intuition.h>
#include <stdio.h>

#include "bars.h"
#include "snapshotgadget.h"
#include "barsmacros.h"
#include "bplib.h"
#include "tuning.h"
#include "proto/tuning_protos.h"
#include "tuningw2.h"
#include "/icons/tuning.h"

#define FILEDUMP 1
#include "debug.h"

UBYTE versiontag[] = "$VER: Alternate Tuning 40.01 (01.03.99)";

#define TOOL_SIZE sizeof(struct TuneTool)

static struct TuningToolMaster master;

extern struct Functions *functions;

/** chicanery section **/
/* this struct is used only to get et into the menus */
static struct _BoGuS_ {
	struct Tuning *next;
	struct String *name;
} bogus;

/* this String * points at 'Equal Temperament' after inittoolmaster */
struct String *et_string = NULL;

#if DBUG
static void dumptuninglist(char *s,struct Tuning *list)
{
	dprintf("%s\n",s);
	for (;list;list=list->next) {
		s=(char *)functions->stringtext(list->name));
		dprintf("tuning @%ld next=%ld %s\n",list,list->next,s);
	}
}
#endif


#if DBUG
static void DumpPitchTable(char *s, struct Tuning *t)

{
	UBYTE i;
	
	dprintf("\n%s PITCH TABLE DUMP\n",s);
	if (t->name) dprintf("%s\n",functions->stringtext(t->name));
	for ( i=0 ; i<128 ; i++ ) {
		dprintf("%3d: %4d %3d %3d %s",i,t->notes[i].midinote,
		         t->notes[i].bendlsb,t->notes[i].bendmsb,
					(i%4)?"  ":"\n");
	}
}
#endif


/* text for bozo requests */
//static const char *BozoTitle = "Alternate Tuning";
static const char *BozoContinue = "Continue";


#define tw (tool->tool.window)

/*==========================================================================
* inittool
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: sets tool defaults
*=========================================================================*/

static void __regargs inittool(struct TuneTool *tool)

{
	tool->tuning = (struct Tuning *)&bogus;
	tool->bendlsb = 0;
	tool->bendmsb = 0x40;
	tool->tool.touched |= TOUCH_INIT;
}

/*============================================================================
* processevent
*
*  INPUT: pointer to a NoteEvent to process
* OUTPUT: none
* RETURN: pointer to an Event
*
* EFFECT: if the event is not a note on or note off, squirt it through.  if
*         the event is a note on, allocate a new NoteEvent for the pitch bend
*         from the tuning, and queue it; substitue the tuning's midinote in
*         the received event.  if  the event is a note off, substitute the
*         tuning's midinote in the event and pass it on.  if the high bit of
*         the midinote number is set, ignore the event because the tuning
*         does not define that note.
*===========================================================================*/

/*** some macros mostly for readability ***/
#define tuningnote(t,n)  ( (t)->tuning->notes[(n)].midinote )
#define notedefined(t,n) ( !(tuningnote(t,n) & 0x80) )
#define validtuning(t)   ( (t)->tuning && ((t)->tuning != (struct Tuning *)(&bogus)) )
/* get bend value from tuning t for note n */
#define bendlsb(t,n)     ( (t)->tuning->notes[(n)].bendlsb )
#define bendmsb(t,n)     ( (t)->tuning->notes[(n)].bendmsb )
#define validnote(t,n)   ( !(bendlsb(t,n) & 0x80) )

static struct Event * __stdargs __saveds processevent(struct Event *event)

{
	struct Event *new;
	struct TuneTool *tool = (struct TuneTool *)event->tool;

	if (!(tool->tool.touched & TOUCH_INIT)) inittool(tool);

	event->tool = event->tool->next;
	if (IsNote(event)) {
		if (validtuning(tool)) {
			if (notedefined(tool,event->byte1) && validnote(tool,event->byte1)) {
				if (IsNoteOn(event)) {
					if ((tool->bendmsb != bendmsb(tool,event->byte2)) || 
				     	 (tool->bendlsb != bendlsb(tool,event->byte1)))  {
						if (new = functions->dupeevent(event)) {
							/* new becomes the note; event carries bend info */
							if (--event->time < 0) { /* ensure bend is before note */
								event->time=0;
								new->time++;
							}
							event->status = MIDI_PBEND;
							event->byte1 = tool->bendlsb = bendlsb(tool,new->byte1);
							event->byte2 = tool->bendmsb = bendmsb(tool,new->byte1);
							event->next = new;
							new->next = NULL;
							new->byte1 = tuningnote(tool,new->byte1); /* stick in tuning's note value */
						} //duped event
					} //new bend
					else { /* same bend; just change note number */
						event->byte1 = tuningnote(tool,event->byte1);
					}
				} //note on
				else if (IsNoteOff(event)) {
					event->byte1 = tuningnote(tool,event->byte1);
				}
			} //note defined
		} //valid tuning
	} //real-time note
	else if (IsAllOff(event)) {
		tool->bendlsb = 0;
		tool->bendmsb = 0x40;
	}
	return(event);
}


/*==========================================================================
* processlist
*
*  INPUT: pointers to tool and list, start, stop
* RETURN: pointer to list
*
* EFFECT: inserts appropriate pitch bends in list
*=========================================================================*/

static struct Event * __stdargs __saveds processlist(struct TuneTool *tool,
                                         struct Event *list, long start, long stop)
{
	struct Event *event = list;
	struct Event *bend;
	struct Event *bendlist = NULL;

	if (!tool->tool.touched) inittool(tool);

	if (validtuning(tool)) {
		while (event) {
			if (IsNoteOn(event)) {
				if (notedefined(tool,event->byte1) && validnote(tool,event->byte1)) {
					if ((tool->bendmsb != bendmsb(tool,event->byte1)) || 
				     	 (tool->bendlsb != bendlsb(tool,event->byte1)) )  {
						if (bend = functions->dupeevent(event)) {
							if (--bend->time < 0) bend->time = 0;
							bend->next = bendlist;
							bendlist = bend;
							bend->status = MIDI_PBEND;
							bend->byte1 = tool->bendlsb = tuningnote(tool,event->byte1);
							bend->byte2 = tool->bendmsb = tuningnote(tool,event->byte2); //tool->tuning->notes[event->value].bendmsb;
						}
					}
					event->byte1 = tuningnote(tool,event->byte1);
				}
			}
			event = event->next;
		}
	}
	if (list && bendlist) {
		list = (struct Event *)functions->List_Cat((struct ListItem *)list,
		                                           (struct ListItem *)bendlist);
		list = (struct Event *)functions->sorteventlist((struct Event *)list);
	}
	/* reset */
	tool->bendlsb=0;
	tool->bendmsb=0x40;
	return(list);
}


/* read and verify */
#define readv(f,d,l) ( functions->fastread((f),(char *)(&(d)),(l)) != -1 )

/*==========================================================================
* readtuning
*
*  INPUT: file pointer, pointer to Tuning
* RETURN: 0 if successful; error code if not
*
* EFFECT: reads a file and puts data in the pitchtable passed
*=========================================================================*/

static long __regargs readtuning(long file, struct Tuning *tuning)

{
	long filesize;
	long size;
	ULONG i;
	char *name, *ptab;
	long nameSize, ptabSize;
	char *dopt;
	long doptSize;
	long error;
	char s[100];

	if (readv(file,i,4)) {
		if (i == ID_TUNE) {
			if (readv(file,filesize,4)) {
				if (filesize >= TOONDATASIZE) {
//dprintf("filesize=%ld\n",filesize);
					filesize -= 8;
					error = 0;
					/* read the file and hang on to the chunks */
					while ( filesize && (error < 4) ) {
						if (readv(file,i,4)) { /* next id */
							if (readv(file,size,4)) { /* chunk size */
								filesize -= 8;
								switch (i) {
									case ID_RATS :
//dprintf("skipping RATS (%ld)\n",size);
										functions->fastseek(file,size,0);
										break;
									case ID_NAME :
										if (!(name = readchunk(file,size))) error = 2;
										nameSize = size;
//dprintf("read NAME (%ld)\n",nameSize);
										break;
									case ID_DOPT :
										if (!(dopt = readchunk(file,size))) error = 3;
										doptSize = size;
//dprintf("read DOPT (%ld)\n",doptSize);
										break;
									case ID_PTAB :
										if (!(ptab = readchunk(file,size))) error = 4;
										ptabSize = size;
//dprintf("read PTAB (%ld)\n",ptabSize);
										break;
									default :
//dprintf("skipping unknown chunk (%ld: %ld bytes)\n",i,size);
										functions->fastseek(file,size,0);
										break;
								}
								filesize -= size;
							} // if (readv(size))
							else error = 5;
						} // if (readv(id))
						else error = 6;
					} // while (filesize)

					if (error < 4) { /* can live without anything but pitch table */
						tuning->rootpitch = *ptab;
						tuning->rootoctave = *(ptab+1);
						memcpy((char *)(&tuning->notes),ptab+2,ptabSize-2);
						if (name) {
							memcpy(s,name,nameSize);
							s[nameSize] = 0;
//dprintf("name=%s\n",s);
							tuning->name = functions->allocstring(s);
						}
						else {
#pragma msg 104 ignore push
							tuning->name = functions->allocstring("Unnamed Tuning");
#pragma msg 104 ignore pop
						}
						if (dopt) {
							tuning->divisions = (char)*(dopt+3);
//dprintf("divisions=%d  ",tuning->divisions);
							tuning->dopt_flags = (char)*(dopt+2);
//dprintf("dopt_flags=%s\n",showbits(tuning->dopt_flags,8));
						}
						else {
							tuning->divisions = 12;
						}
						error = 0; /* we been compensating */
					} //if (error...

					/* clean up readchunk()'s allocations */
					if (ptab) functions->myfree(ptab,ptabSize);
					if (name) functions->myfree(name,nameSize);
					if (dopt) functions->myfree(dopt,doptSize);
				}
				else error = 100; /* filesize too small */
			}
			else error = 101; /* no filesize to read */
		}
		else error = 102; /* wrong id */
	}
	else error = 103; /* can't read id */

	return(error);
}


/*==========================================================================
* tuninginlist
*
*  INPUT: pointer to struct Tuning
* RETURN: pointer to tuning with same name, or NULL if no match
*
* EFFECT: looks for a tuning with a duplicate name in the list of loaded
*         tunings
*=========================================================================*/

static struct Tuning * __regargs tuninginlist(tuning)

struct Tuning *tuning;

{
	char name[100]; /*?? long enough ?*/
	struct Tuning *list;

	if (tuning && tuning->name) {
		strcpy(name,(char *)(functions->stringtext(tuning->name)));
		for ( list=master.list->next ; list ; list=list->next ) {
			if ( strcmp(name,(char *)(functions->stringtext(list->name))) == 0 )
				return(list); /* pointer to loaded match */
		}
	}
	/*else*/
	return(NULL); /* no match in list */
}

/*==========================================================================
* loadtuningfile
*
*  INPUT: none
* RETURN: 1 if successful; 0 if not
*
* EFFECT: gets a filename and calls readtuning()
*         as with savetuningfile(), this function is the interface with the
*         real guts.  modify here to customize for a Tool or Accessory.
*=========================================================================*/

static struct Tuning * __stdargs __saveds loadtuningfile()

{
	long file;
	long error = 0;
	struct Tuning *tail;
	struct Tuning *new;
#pragma msg 104 ignore push
	static char filename[100] = "";
#pragma msg 104 ignore pop
	char text[140]; /* for bozo request if can't load tuning file */

#pragma msg 104 ignore push
	functions->FileName(filename,"Open Tuning File","TUN",functions->screen,
	                    FILES_OPEN|FILES_TYPE,0,0);
#pragma msg 104 ignore pop
	if (filename[0]) {
		functions->openwait();
		if (file = functions->fastopen(filename,MODE_OLDFILE)) {
			/* allocate and clear memory for a new tuning table */
			if (new = (struct Tuning *)
			             functions->myalloc(TUNING_NORAT_SIZE,MEMF_CLEAR)) {
				memset((char *)new,TUNING_NORAT_SIZE,0);
				if (error = readtuning(file,new)) {
//dprintf("readtuning returned error=%ld\n",error);
					functions->myfree((char *)new,TUNING_NORAT_SIZE);
					new = NULL;
				}
				else {
					if (tuninginlist(new)) { /* avoid same-named tunings */
#pragma msg 104 ignore push
						if (!functions->areyousure("duplicate a loaded tuning?")) {
#pragma msg 104 ignore pop
							if (new->name) functions->freestring(new->name);
							functions->myfree((char *)new,TUNING_NORAT_SIZE);
							new = NULL;
						}
					}
					if (!error) {
						/* link new tuning */
						new->next = NULL;
						if (tail = (struct Tuning *)listend((APTR)master.list)) 
							tail->next = new;
						else { /* this is very bad news */
							error=2;
						}
					}
				}
			}
			functions->fastclose(file);
		}
		functions->closewait();
	}
	if (error) {
#pragma msg 104 ignore push
		sprintf(text,"Can't open the file\n%s!",filename);
//		BozoRequest(NULL,BozoTitle,text,0,-1,-1,1,BozoContinue,NULL);
		functions->msgbox(text,BozoContinue,NULL,NULL,1,BP_LGREY,BP_RED,BP_BLUE);
#pragma msg 104 ignore pop
		if (new) {
			if (new->name) functions->freestring(new->name);
			functions->myfree((char *)new,TUNING_NORAT_SIZE);
			new = NULL;
		}
	}
//else DumpPitchTable("Freshly loaded",new);
	return(new);
}

/* some ScrollingPopUpMenu() arguments for consistency between select & remove */
#define POPUP_WIDTH 160
#define POPUP_PEN 6
#define POPUP_Y   8
#define POPUP_MAXDISP 5

/*============================================================================
* drawitem
*
*  INPUT: pointer to RastPort to draw in, pointer to item to draw, TopEdge
* RETURN: none
*
* EFFECT: displays name of a tuning for ScrollingPopUpMenu()
*=========================================================================*/

static void __stdargs __saveds drawitem(struct RastPort *rp,struct Tuning *item,
                                        short y)
{
	printString(rp,2,y+6,POPUP_PEN,0,JAM1,item->name);
}

/*==========================================================================
* selecttuning
*
*  INPUT: pointer to tool
* OUTPUT: none
* RETURN: pointer to the tuning selected
*
* EFFECT: displays a menu of loaded tunings, gets user selection, points
*         tool->tuning at selection
*=========================================================================*/

static struct Tuning *selecttuning(struct TuneTool *tool)

{
	struct Tuning *item;
	
#pragma msg 88 ignore push
	item = (struct Tuning *)functions->ScrollingPopUpMenu(functions->screen,
	               (APTR)master.list,drawitem,POPUP_Y,POPUP_WIDTH,POPUP_MAXDISP);
#pragma msg 88 ignore pop
	if (item) {
		tool->tuning = item;
		tool->tool.touched |= TOUCH_INIT;
	}
	return(item);
}

/*==========================================================================
* deletetuning
*
*  INPUT: pointer to tuning
* RETURN: none
*
* EFFECT: if tuning is in master.list, it is freed and the list adjusted
*         accordingly
*=========================================================================*/

static void __regargs deletetuning(struct Tuning *t)

{
	struct Tuning *prev;
	struct Tuning *tuning;

	if (t && (t != (struct Tuning *)(&bogus))) {
		/* find it */
		for ( prev=master.list, tuning=master.list->next ;
	   	   tuning && (tuning != t) ; 
		   	prev=tuning, tuning=tuning->next )
			;
		/* get rid of it */
		if (tuning == t) { /* find it? */
			prev->next = t->next;
			if (t->name) functions->freestring(t->name);
			functions->myfree((char *)t,TUNING_NORAT_SIZE);
		}
	}
}

/*==========================================================================
* removetuning
*
*  INPUT: pointer to tool
* OUTPUT: none
* RETURN: none
*
* EFFECT: removes a selected tuning table from the toolmaster's list.  scans
*         tracklist to locate other tunetools that are using the selected
*         tuning.  tools using the selected tuning have their tuning table
*         pointer set to &bogus.
*=========================================================================*/

static void __regargs removetuning(struct TuneTool *tool)

{
	struct Tuning *item;
	struct TuneTool *t;
	struct Track *track;
	UBYTE inuse;
	char text[60];

#pragma msg 88 ignore push
	item = (struct Tuning *)functions->ScrollingPopUpMenu(functions->screen,
				(APTR)master.list->next,drawitem,POPUP_Y,POPUP_WIDTH,POPUP_MAXDISP);
#pragma msg 88 ignore pop
	if ( !item ) return;
	/* else */
	/* check for an instance of this tool using the tuning to be removed */
	inuse = 0;
	for ( track=functions->tracklist ; track && !inuse ; track=track->next ) {
		for ( t=(struct TuneTool *)track->toollist ; t ; 
		               t=(struct TuneTool *)t->tool.next ) {
			if ( (t->tool.toolid == ATUN_TOOLID) && (t->tuning == item) ) {
				inuse = 1;
				break;
			}
		}
	}
	if (inuse) { /* used by some tool */
		sprintf(text,"remove %ls? It is in use.",functions->stringtext(item->name));
		if (!functions->areyousure(text)) return;
	}
	/*** expunge that puppy ***/
	/** reset all pointers to the tuning **/
	for ( track=functions->tracklist ; track ; track=track->next ) {
		for ( t=(struct TuneTool *)track->toollist ; t ; 
		      							t=(struct TuneTool *)t->tool.next ) {
			if ( (t->tool.toolid == ATUN_TOOLID) && (t->tuning == item) ) {
				t->tuning = (struct Tuning *)&bogus; /* revert to equal temperament */
			}
		}
	}
	deletetuning(item);
}

/*==========================================================================
* showtuningname
*
*  INPUT: pointer to tool edit window, pointer to tool, pen number
* RETURN: none
*
* EFFECT: displays the name of a tuning
*=========================================================================*/

#define TUNINGNAME_X 79 //80
#define TUNINGNAME_Y 23
#define TUNINGNAME_MAXCHAR 25

static void __regargs showtuningname(struct TuneTool *tool, short pen)

{
/*** WHAT ASSWIPE made tuning.name a String * (@#*&!!@) ***/
	char name[TUNINGNAME_MAXCHAR+1];

	if (tool->tuning->name) {
		sprintf(name,"%-25.25s",functions->stringtext(tool->tuning->name));
		name[TUNINGNAME_MAXCHAR] = '\0';
		SetRPort(tw->RPort,TUNINGNAME_X,TUNINGNAME_Y,pen,0,JAM2);
		Text(tw->RPort,name,TUNINGNAME_MAXCHAR);
	}
}

/*==========================================================================
* FindAccessory
*
*  INPUT: pointer to string (name of accessory)
* RETURN: pointer to Accessory
*
* EFFECT: searches functions->accesslist for name accessory
*=========================================================================*/

struct Accessory *FindAccessory(const char *name)

{
	struct Accessory *a = functions->accesslist;

	for ( ; a ; a = a->next)  if (strcmp(a->name,name) == 0) break;
	return(a);
}


/*** some error codes for import/export (note that they map to errormsg[]) ***/

/* not enough */
#define MSG_ERR_MEMORY 1
/* weird communication */
#define MSG_ERR_GARBLE 2
/* tonleitermeister not found, or not running */
#define MSG_ERR_NOACCESS 3
/* tonleitermeister said bite me -- handle in eartask()? */
#define MSG_ERR_UNRELIABLE 4

static const char * errormsg[] = {
	"Unable to transfer tuning;\nprobably not enough memory.",
	"Unable to communicate with the\nTonleitermeister accessory.",
	"The Tonleitermeister accessory must\nbe running to transfer a tuning.",
	"Tonleitermeister did not accept the tuning." 
};

/*==========================================================================
* ExportTuning
*
*  INPUT: pointer to tool, pointer to MakeTuneAccessory
* RETURN: none
*
* EFFECT: hangs out until Tonleitermeister either copies the tool's currently
*         selected tuning or gives up trying to copy.
*=========================================================================*/

static void ExportTuning(struct TuneTool *tool)

{
	struct TuneMessage *msg;
	struct MakeTuneAccessory *acc;
	UBYTE error;

	if (!master.mp) return; // couldn't create port; can't export
	if (acc = (struct MakeTuneAccessory *)FindAccessory("Tonleitermeister")) {
//dprintf("exp() acc=%X task=%X\n",acc,acc->task);
		if (!(acc->running)) {
			/* start it, and wait for it */
#pragma msg 154 ignore push
			if ( functions->doscall(acc->acc.open) ) {
#pragma msg 154 ignore pop
/*				while (!(acc->running)) WaitTOF();
				WaitTOF(); /* give it a chance to initialize itself */
*/			}
			else {
/*				BozoRequest(NULL,BozoTitle,"Can't start Tonleitermeister!",
				            0,-1,-1,1,BozoGadgetText,NULL);
*/
				functions->msgbox("Can't start Tonleitermeister accessory!\nOut of memory?",
				                   BozoContinue,NULL,NULL,1,BP_LGREY,BP_RED,BP_BLUE);
				return;
			}
		}
	}
	else {
/*		BozoRequest(NULL,BozoTitle,
         "You must install the Tonleitermeister\naccessory to edit a tuning.",
	               0,-1,-1,1,BozoGadgetText,NULL);
*/
		functions->msgbox("The Tonleitermeister accessory must\n be installed to edit a tuning.",
		                  BozoContinue,NULL,NULL,1,BP_LGREY,BP_RED,BP_BLUE);
		return;
	}

	if (acc->mp) {
		if (msg = (struct TuneMessage *)
                functions->myalloc(sizeof(struct TuneMessage),MEMF_CLEAR)) {
//if (master.mplock) dprintf("exp() mplock\n");
			while (master.mplock) WaitTOF();
			master.mplock = 1;
			master.mp->mp_SigTask = FindTask(NULL); /* be sure to get signalled */
			msg->message.mn_ReplyPort = master.mp;
			msg->command = CMD_EDIT;
			msg->tuning = tool->tuning;
//dprintf("exp() put EDIT @ %X tuning=%X\n",msg,msg->tuning);
			PutMsg(acc->mp,(struct Message *)msg);
			WaitPort(master.mp); /* wait to get a reply */
			if (msg = (struct TuneMessage *)GetMsg(master.mp)) {
				if (msg->command == CMD_ACK) { /* ok */
//dprintf("exp() got ACK\n");
					error = 0;
					functions->myfree((char *)msg,sizeof(struct TuneMessage));
				}
				else if (msg->command == CMD_BITEME) { /* won't take it */
//dprintf("exp() got BITEME\n");
					error = MSG_ERR_UNRELIABLE;
					functions->myfree((char *)msg,sizeof(struct TuneMessage));
				}
				else {
					if (msg->message.mn_ReplyPort != master.mp) {
						msg->command = CMD_BITEME;
						msg->tuning = NULL;
//dprintf("exp() says BITEME @ %X to %X\n",msg,msg->message.mn_ReplyPort);
						ReplyMsg((struct Message *)msg);
					}
					else {
//dprintf("exp() got back %X\n",msg);
						functions->myfree((char *)msg,sizeof(struct TuneMessage));
					}
				}
			}
			master.mplock = 0;
		}
		else error=MSG_ERR_MEMORY;
	}
	else error=MSG_ERR_GARBLE;

	if (error) {
//dprintf("Export() error = %d\n",error);
//		BozoRequest(NULL,BozoTitle,errormsg[error-1],0,-1,-1,0,BozoGadgetText,NULL);
		functions->msgbox(errormsg[error-1],BozoContinue,NULL,NULL,1,BP_LGREY,BP_RED,BP_BLUE);
	}
}


/*==========================================================================
* ImportTuning
*
*  INPUT: none
* RETURN: pointer to Tuning, or NULL
*
* EFFECT: gets a tuning map from Tonleitermeister
*=========================================================================*/

static struct Tuning * ImportTuning(void)

{
	struct MakeTuneAccessory *acc;
	struct TuneMessage *msg;
	struct Tuning *tuning = NULL;
	struct Tuning *tail;
	UBYTE error = 0;

	if (!master.mp) return(NULL); // couldn't create port; can't import
	if (acc = (struct MakeTuneAccessory *)FindAccessory("Tonleitermeister")) {
		if (!(acc->running)) error = MSG_ERR_NOACCESS; /* user is an asswipe */
		else {
			if (acc->mp) {
				if (msg = (struct TuneMessage *)
                   functions->myalloc(sizeof(struct TuneMessage),MEMF_CLEAR)) {
//if (master.mplock) dprintf("imp() mplock\n");
					while (master.mplock) WaitTOF(); /* wait to use msgport */
					master.mplock = 1;
					master.mp->mp_SigTask = FindTask(NULL); /* be sure to get signalled */
					msg->message.mn_ReplyPort = master.mp;
					msg->command = CMD_GIMMEE;
					PutMsg(acc->mp,(struct Message *)msg);
					WaitPort(master.mp); /* wait to get the pointer to the tuning */
					if (msg = (struct TuneMessage *)GetMsg(master.mp)) {
						if (msg->command == CMD_TAKEIT) {
							/* got a pointer; copy the tuning */
							if (tuning = (struct Tuning *)
							         functions->myalloc(TUNING_NORAT_SIZE,MEMF_CLEAR)) {
								Forbid();
								memcpy((char *)tuning,(char *)msg->tuning,TUNING_NORAT_SIZE);
								tuning->name = functions->allocstring(functions->stringtext(msg->tuning->name));
								Permit();
/*
								tuning->next = NULL;
								if (tail = (struct Tuning *)listend((APTR)master.list)) 
									tail->next = tuning;
*/
							}
							else error=MSG_ERR_MEMORY;
							/* send an ack to free the lock on the tuning */
							msg->message.mn_ReplyPort = NULL; /* always freed by receiver */
							msg->command = CMD_ACK;
							PutMsg(acc->mp,(struct Message *)msg);
						}
						else {
//dprintf("ATun got WEIRD SHIT: command=0x%8X tuning=0x%8X\n",msg->command,msg->tuning);
							error=MSG_ERR_GARBLE;
							functions->myfree((char *)msg,sizeof(struct TuneMessage));
						}
/******
						if (msg->message.mn_ReplyPort || (msg->command == CMD_ACK)) {
dprintf("ATun replying 0x%8X\n",msg);
							ReplyMsg((struct Message *)msg); /* ack freed by accessory */
						}
						else {
//dprintf("ATun freeing msg @ 0x%8X\n",msg);
							functions->myfree((char *)msg,sizeof(struct TuneMessage));
						}
*******/
					} // if (msg=GetMsg...
					master.mplock = 0;

				} // if (msg...
				else {
					error=MSG_ERR_MEMORY;
				}
			}
			else {
				error=MSG_ERR_GARBLE;
			} // if (acc->mp)
		}
	} // if (acc)
	else {
		error = MSG_ERR_NOACCESS;
	}

	if (error) {
//dprintf("Import() error = %d\n",error);
//		BozoRequest(NULL,BozoTitle,errormsg[error-1],0,-1,-1,0,BozoGadgetText,NULL);
		functions->msgbox(errormsg[error-1],BozoContinue,NULL,NULL,1,BP_LGREY,BP_RED,BP_BLUE);
	}
	else {
		if (tuninginlist(tuning)) {
/*			if (BozoRequest(NULL,BozoTitle,
			                "Do you really want to\nimport a duplicate tuning?",
								 0,-1,-1,1,"Yes","No") != BR_YES) {
*/
			if (functions->msgbox("Do you really want to\nimport a duplicate tuning?"
			    ," Ja ","Nein",NULL,1,BP_LGREY,BP_RED,BP_BLUE) != 0) {
				functions->freestring(tuning->name);
				deletetuning(tuning);
				tuning=NULL;
			}
		}
	   /* link it if it still exists */
      if (tuning) {
		   tuning->next = NULL;
   		if (tail = (struct Tuning *)listend((APTR)master.list)) 
				tail->next = tuning;
		}
   }
	return(tuning);
}


/** color to write tuning name **/
#define SHOW_PEN 1

/*==========================================================================
* edittool
*
*  INPUT: pointer to the tool that called
* OUTPUT: none
* RETURN: none
*
* EFFECT: mess with gadgetry and stuff
*===========================================================================*/

static void __stdargs __saveds edittool(struct TuneTool *tool)

{
	struct IntuiMessage *message;
	long class; 
	long code;
	struct Gadget *gadget;
	struct NewWindow *nw;
	UBYTE refresh = 1;

	if (!(tool->tool.touched & TOUCH_INIT)) inittool(tool);
//	nw = OpenEmbossedWindowSlop(&NewWindowStructure1,&tool->tool,0,WTITLE,20);
	nw = NewOpenEmbossedWindow(&NewWindowStructure1,&tool->tool,0,WTITLE);
	if (!nw) return;
	SetWindowTitles(tw,(UBYTE *)-1,(UBYTE *)"AlternateTuning  v1.4  1991-99 Another Option");
//	AddSnapshotGadget(tw);
	functions->EmbossOn(tw,SELECT,0);
	EmbossOnGList(tw,&Gadget2,GCOUNT-1,1);
	DrawBevelBox(tw->RPort,76,15,281,25,-1,-1,1); /* frame for selected tuning */
	SetMenuStrip(tw,&Menu1);
	/* disable importing if msgport does not exist */
	functions->EnableEmbossed(tw,GRAB,(master.mp != NULL));

	for (;;) {
//dumptuninglist("\nedittool()",master.list);
		if (refresh) {
			functions->EnableEmbossed(tw,REMOVE,(master.list->next != NULL));
			functions->EnableEmbossed(tw,SELECT,(master.list->next != NULL));
			functions->EnableEmbossed(tw,EDIT,(tool->tuning != master.list) && master.mp);
			showtuningname(tool,SHOW_PEN);
			refresh = 0;
		}

		message = (struct IntuiMessage *)functions->GetIntuiMessage(tw);
		class = message->Class;
		code = message->Code;
		gadget = (struct Gadget *)message->IAddress;
		ReplyMsg((struct Message *)message);
		class = functions->SystemGadgets(tw,class,gadget,code);

		if (class == CLOSEWINDOW)
			break;
		else if (class == GADGETUP) {
			if (gadget->GadgetID == LOAD) {
#pragma msg 154 ignore push
				if (class = (long)functions->doscall(loadtuningfile)) {
#pragma msg 154 ignore pop
					tool->tuning = (struct Tuning *)class;
					refresh=1;
				}
			}

			else if (gadget->GadgetID == EDIT) { /* try to export current tuning */
				ExportTuning(tool);
			}

			else if (gadget->GadgetID == GRAB) { /* try to import a tuning */
				if (class = (long)ImportTuning()) {
					tool->tuning = (struct Tuning *)class;
					refresh = 1;
				}
			}
		}
		else if (class == GADGETDOWN) {
			if (gadget->GadgetID == SELECT) {
				selecttuning(tool);
				refresh=1;
			}
			else if (gadget->GadgetID == REMOVE) {
				removetuning(tool);
				refresh=1;
			}
		}
		else if (class == MENUPICK) {
			if (ITEMNUM(code) == MENU_LOAD) {
#pragma msg 154 ignore push
				if (class = (long)functions->doscall(loadtuningfile)) {
#pragma msg 154 ignore pop
					tool->tuning = (struct Tuning *)class;
					refresh=1;
				}
			}
/*
			else if (ITEMNUM(code) == MENU_SAVE) {
				/* save a tuning file */
			}
*/
			else if (ITEMNUM(code) == MENU_EDIT) {
				ExportTuning(tool);
			}
			else if (ITEMNUM(code) == MENU_GRAB) {
				/* try to get on from Tonleitermeister */
				if (class = (long)ImportTuning()) {
					tool->tuning = (struct Tuning *)class;
					refresh=1;
				}
			}
/*
			else if (ITEMNUM(code) == MENU_REMOVE) {
				/* try to remove a tuning -- IF you really want to deal with
				   managing all the loaded tunings in an Intuition menu */
				refresh=1;
			}
*/
		}
	}

	EmbossOffGList(tw,&GadgetList1,GCOUNT);
	NewCloseEmbossedWindow(&tool->tool,nw);
}

/*==========================================================================
* removetool
*
*  INPUT: none
* RETURN: none
*
* EFFECT: frees allocated memory, replies or frees outstanding messages, and
*         deletes message port.
*=========================================================================*/

/*!! need to wait on a master.mplock !!*/

static void __stdargs __saveds removetool()

{
	struct Tuning *t;
	struct TuneMessage *msg;

	while (master.mplock) WaitTOF();
	if (master.list) {
		for ( t=master.list->next ; t ; t=t->next ) {
			deletetuning(t);
		}
		if (et_string) functions->freestring(et_string);
		bogus.name = NULL;
		master.list = NULL;
		if (master.mp) {
			while (msg = (struct TuneMessage *)GetMsg(master.mp)) {
				if (msg->message.mn_ReplyPort) {
					msg->command = CMD_BITEME;
					ReplyMsg((struct Message *)msg);
				}
				else functions->myfree((char *)msg,sizeof(struct TuneMessage));
			}
			DeletePort(master.mp);
			master.mp = NULL;
		}
	}
}

/*==========================================================================
* savesize
*
*  INPUT: pointer to tool
* RETURN: bytes required to save the tool and its Tuning
*
* EFFECT: checks and adds
*=========================================================================*/

static long __stdargs __saveds savesize(tool)

struct TuneTool *tool;

{
	if (validtuning(tool)) {
		/* a user-defined tuning selected */
		return((long)(TOOL_SIZE + TOONDATASIZE +
		       strlen(functions->stringtext(tool->tuning->name))) );
	}
	else {
		/* equal temperament -- no pitch table to save */
		return(TOOL_SIZE);
	}
}

/*==========================================================================
* loadtool
*
*  INPUT: file id, size of chunk to read
* RETURN: pointer to allocated tool, or NULL
*
* EFFECT: allocates and initializes a tool; reads a pitch table
*=========================================================================*/

static struct Tool * __stdargs __saveds loadtool(file,size)

long file;
long size;

{
	struct TuneTool *tool;
	struct Tuning *tuning;
	char error = 0;

	if ( tool = (struct TuneTool *)
	            functions->myalloc(sizeof(struct TuneTool),MEMF_CLEAR) ) {
		functions->fastread(file,(char *)tool,TOOL_SIZE);
		size -= TOOL_SIZE;
		inittool(tool);
		if (size) { /* pitch table was written */
			if ( tool->tuning = (struct Tuning *)functions->myalloc
			                             (sizeof(struct Tuning),MEMF_CLEAR) ) {
				if (readtuning(file,tool->tuning) == 0) { /* success */
					size -= savesize(tool) - TOOL_SIZE;
					if (!(tuning = tuninginlist(tool->tuning))) {
						/* add this one to the master list */
						if (tuning = (struct Tuning *)listend((APTR)master.list)) {
							tuning->next = tool->tuning;
							tool->tuning->next = NULL;
						}
						else error=1; /* this means big trouble */
					}
					else { /* it exists; don't duplicate */
						if (tool->tuning->name) functions->freestring(tool->tuning->name);
						functions->myfree((char *)tool->tuning,sizeof(struct Tuning));
						tool->tuning = tuning;
					}
				}
				else error=2;
			}
			else error=3;
		}
	}
	if (size) { /* leftovers? */
		functions->fastseek(file,size,0);
	}

	if (error) { /* clean it up */
		if (tool) {
			if (tool->tuning) {
				if (tool->tuning->name) {
					functions->freestring(tool->tuning->name);
				}
				functions->myfree((char *)tool->tuning,sizeof(struct Tuning));
			}
			functions->myfree((char *)tool,sizeof(struct TuneTool));
			tool = NULL;
		}
	}

	return((struct Tool *)tool);
}

/*==========================================================================
* writepitchtable
*
*  INPUT: file pointer, pointer to pitch table, mask flag
* RETURN: 1 if successful, 0 if not
*
* EFFECT: writes contents of the pitch table to file.  if mask==1, midinotes
*         are written straight from the table.  if mask==0, only valid MIDI
*         data is written from the table.
*=========================================================================*/

static long __inline writepitchtable(file,ptab,mask)

long file;
struct Tuning *ptab;
UBYTE mask;

{
	ULONG i;
	struct PitchDef note;

	i = ID_PTAB;
	if (writev(file,i,4)) {
		i = sizeof(ptab->notes)+2;
		if (writev(file,i,4)) {
			/* roots are useless, but make the read easier */
			writebyte(file,ptab->rootpitch);
			writebyte(file,ptab->rootoctave);
			if (mask) { /* spit out masked pitchtable as is */
				if ( functions->fastwrite(file,(char *)(&ptab->notes),
				                            sizeof(ptab->notes)) != -1 ) {
					i = 1L;
				}
			}
			else { /* dribble notes and bends out unmasked */
				for (i=0 ; i<128 ; i++) {
					memcpy((char *)&note,(char *)(&ptab->notes[i]),sizeof(struct PitchDef));
					note.midinote &= 0x7F;
					if (!writev(file,note,3)) break;
				}
				i = (i == 128);
			}
		}
		else i=0;
	}
	else i=0;

	return((long)i);
}

/*==========================================================================
* writetuning
*
*  INPUT: file, pointer to pitch table
* RETURN: positive long if successful; 0 if not
*
* EFFECT: writes it to file. if a display option is -1, it is set to 0.
*         if ratios==-1, ptab->ratios[][] will not be written.
*=========================================================================*/

static long writetuning(file,ptab)

long file;
struct Tuning *ptab;

{
	ULONG i;
	ULONG n;
	char *node;

	i = ID_TUNE;
	if (writev(file,i,4)) {
		node = (char *)functions->stringtext(ptab->name);
		n = __builtin_strlen(node) + TOONDATASIZE; /* file size */
		if (writev(file,n,4)) {
			i = ID_NAME;
			writev(file,i,4);
			n -= TOONDATASIZE;
			writev(file,n,4); /* nameSize */
			if (functions->fastwrite(file,node,n) != -1) {	/* name */
				i = ID_DOPT;
				writev(file,i,4);
				i = 4;
				writev(file,i,4);
				/* stash dopt bits in byte2; divisions in byte1 */
				i = (ptab->dopt_flags << 8) | ptab->divisions;
				if writev(file,i,4) {	/* display options */
					i = writepitchtable(file,ptab,ptab->dopt_flags & DOPT_MASK);
				}
				else i=0;
			}
			else i=0;
		}
		else i=0;
	}
	else i=0;
	
	return((long)i);
}

/*==========================================================================
* savetool
*
*  INPUT: file id, pointer to tool to save
* RETURN: 0 if successful; 1 otherwise
*
* EFFECT: writes a tool's data to a file.
*=========================================================================*/

static long __stdargs __saveds savetool(file,tool)

long file;
struct TuneTool *tool;

{
	long size;
	
	functions->fastwrite(file,(char *)(&tool->tool.toolmaster->toolid),4);
	size = savesize(tool);
	functions->fastwrite(file,(char *)(&size),4);
	/* write the tool data -- even the useless stuff */
	functions->fastwrite(file,(char *)tool,TOOL_SIZE);
	if (validtuning(tool)) {
		if (!writetuning(file,tool->tuning)) {
			return(1);
		}
	}
	return(0); /* ok */
}

/*==========================================================================
* inittoolmaster
*
*  INPUT: none
* RETURN: pointer to tool master struct
*
* EFFECT: initializes.  in addition to the usual stuff, this function also
*         sets up the bogus tuning stucture for equal temperament, and 
*         points the tool master list field at the bogus structure.
*=========================================================================*/

struct ToolMaster * __stdargs __saveds inittoolmaster(void)

{
	memset((char *)&master,0,sizeof(struct TuningToolMaster));
	master.master.toolsize = sizeof(struct TuneTool);
	master.master.toolid = ATUN_TOOLID;
	master.master.edittool = edittool;
	master.master.processevent = processevent;
	master.master.processlist = processlist;
	master.master.removetool = removetool;
	master.master.savesize = savesize;
	master.master.savetool = savetool;
	master.master.loadtool = loadtool;
	master.master.image = &TuningIcon;
	strcpy(master.master.name,"Alternate Tuning");
	master.master.tooltype = TOOL_NORMAL;

	/*** peculiar chicanery stuff ***/
#pragma msg 104 ignore push
	if (et_string = functions->allocstring("Equal Temperament")) {
#pragma msg 104 ignore pop
		master.list = (struct Tuning *)&bogus;
		master.list->next = NULL;
		bogus.name = et_string;
		/* set up the port to talk to Tonleitermeister */
		if (!(master.mp = CreatePort(TOOL_PORT_NAME,0))) {
//dprintf("ATun can't create port\n");
/*			BozoRequest(NULL,BozoTitle,
			            "Cannot create message port\nfor Tonleitermeister.",
					      0,-1,-1,1,BozoGadgetText,NULL);
*/
		functions->msgbox("AlternateTuning can't create a message port\nfor the Tonleitermeister accessory.\nEdit will be disabled.",
		                  BozoContinue,NULL,NULL,1,BP_LGREY,BP_RED,BP_BLUE);
		}
		master.mplock = 0;
	}
	else {
		functions->msgbox("Not enough memory to initialize\nAlternateTuning.ptool\n",
		                  BozoContinue,NULL,NULL,1,BP_LGREY,BP_RED,BP_BLUE);
		return(NULL);
	}

	return((struct ToolMaster *)&master);
}
